from typing import Any, Self
from lptp import exceptions
from .types import FieldType


class DataField:
    name: int
    type: FieldType
    data: Any

    def __init__(self, name: int, data: Any) -> None:
        if not isinstance(name, int):
            raise TypeError("name should be instance of int")
        if name > 0xFF or name < 0:
            raise ValueError("name should be within 0xFF")
        
        data_type = FieldType.by_datatype(type(data))
        if not data_type:
            raise TypeError(f"{type(data).__name__} is invalid FieldType")
        
        self.name = name
        self.type = data_type
        self.data = data

    def to_bytes(self) -> bytes:
        name = self.name.to_bytes()
        data = self.type.to_bytes(self.data)
        field_type = self.type.raw_type.to_bytes()
        field_length = len(data).to_bytes(2)
        return name + field_type + field_length + data
    
    def __repr__(self) -> str:
        name = self.name
        data = self.data
        date_type = self.type
        return f"<{type(self).__name__} {name=} {data=} {date_type=}>"

    @classmethod
    def from_bytes(cls, name: int, data: bytes, field_type: FieldType) -> Self:
        if not isinstance(field_type, FieldType):
            raise ValueError(f"field_type should be instance of FieldType")
        return cls(name, field_type.from_bytes(data))
    

class SingleDataField(DataField):

    def __init__(self, data: Any) -> None:
        super().__init__(0, data)

    def to_bytes(self) -> bytes:
        data = self.type.to_bytes(self.data)
        field_type = self.type.raw_type.to_bytes()
        return field_type + data
    
    def __repr__(self) -> str:
        data = self.data
        return f"<{type(self).__name__} {data=}>"
    
    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        if not isinstance(data, bytes):
            raise ValueError(f"data should be instance of bytes")
        if len(data) < 2:
            raise exceptions.DataException("Cannot convert data")
        
        field_type = FieldType.by_rawtype(data[0])
        if not field_type:
            raise exceptions.DataException("Invalid field_type")
        return cls(field_type.from_bytes(data[1:]))


class UnopenedConnection(Exception):
    def __init__(self) -> None:
        super().__init__("Open connection before send packets")

class ClosedConnecton(Exception):
    def __init__(self, message: str = "Connection already closed") -> None:
        super().__init__(message)

class AuthException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)

class ProcedureError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


from typing import Any, Callable, Coroutine, List, Optional
import inspect

from lptp.exceptions.data import DataException
from lptp.types.pds.pds import ProcedureDataStructure
from lptp.types.pds.types import FieldType


class Procedure:
    coro: Callable[..., Coroutine]
    name: Optional[int]
    subtype: int

    def __init__(self, coro: Callable[..., Coroutine], name: Optional[int], subtype: int) -> None:
        if name:
            if not isinstance(name, int):
                raise TypeError("name should be instance of int")
            if name < 1 or name > 255:
                raise ValueError("name should be within 1-0xFF")
        if not isinstance(subtype, int):
            raise TypeError("name should be instance of int")
        if subtype < 0 or subtype > 255:
            raise ValueError("subtype should be within 0xFF")
        
        self.coro = coro
        self.name = name
        self.subtype = subtype
        self.get_fields() # Check for errors

        if not self.get_return():
            raise TypeError(f"{self.get_return_datatype().__name__} is invalid FieldType")

    async def call(self, pds: ProcedureDataStructure) -> Any:
        if not self.validate_fields(pds):
            raise DataException("Invalid fields")
        res = await self.coro(*pds.get_params())
        if not isinstance(res, self.get_return_datatype()):
            raise TypeError(f"Procedure returned invalid response ({type(res).__name__})")
        return res

    def validate_fields(self, pds: ProcedureDataStructure) -> bool:
        fields = self.get_fields()
        valid_pds_fields = []
        for field in fields:
            pds_field = next(filter(lambda f: f.name == field.name, pds.fields), None)
            if not pds_field:
                if field.have_default: continue
                return False
            if not field.field_type.validate_field(pds_field):
                return False
            valid_pds_fields.append(pds_field)
        if len(valid_pds_fields) != len(pds.fields):
            return False
        return True

    def get_func_name(self) -> str:
        return self.coro.__name__

    def get_return_datatype(self) -> type:
        return inspect.getfullargspec(self.coro).annotations.get('return', None)

    def get_return(self) -> Optional[FieldType]:
        return FieldType.by_datatype(self.get_return_datatype())

    def get_fields(self) -> List["ProcedureParam"]:
        spec = inspect.getfullargspec(self.coro)
        defaults = dict(zip(spec.args[-len(spec.defaults):], spec.defaults)) if spec.defaults else {}
        params = {
            name: {
                "type": spec.annotations.get(name, Any)
            }
            for name in self.coro.__annotations__
            if name != "return"
        }
        for k,v in defaults.items():
            params[k]['default'] = v
        return [
            ProcedureParam(i, name, info['type'], 'default' in info, info.get('default', None))
            for i, (name, info) in enumerate(params.items())
        ]
    
    def __repr__(self) -> str:
        name = self.name
        subtype = self.subtype
        coro = self.coro
        return f"<{type(self).__name__} {name=} {subtype=} {coro=}>"


class ProcedureParam:
    name: int
    varname: str
    field_type: FieldType
    have_default: bool
    default: Any

    def __init__(self, name: int, varname: str, type_: type, have_default: bool = False, default: Any = None) -> None:
        if not isinstance(name, int):
            raise TypeError("name should be instance of int")
        if not isinstance(varname, str):
            raise TypeError("varname should be instance of str")
        if not isinstance(have_default, bool):
            raise TypeError("have_default should be instance of bool")
        
        field_type = FieldType.by_datatype(type_)
        if not field_type:
            raise TypeError(f"{type_.__name__} is invalid FieldType")
        
        self.name = name
        self.varname = varname
        self.field_type = field_type
        self.have_default = have_default
        self.default = default

    def __repr__(self) -> str:
        name = self.name
        varname = self.varname
        field_type = self.field_type
        have_default = self.have_default
        default = self.default
        return f"<{type(self).__name__} {name=} {varname=} {field_type=} {have_default=} {default=}>"


from typing import Optional, Self
import asyncio

from lptp.types.packets import Packet, ServerPacket
from lptp.exceptions import UnopenedConnection, ClosedConnecton, AuthException
from lptp.types.packets import AuthPacket, AuthType, StatusCode, ExitPacket
from lptp.types.packets.procedure import ProcedurePacket
from lptp.types.pds.pds import ProcedureDataStructure


class LPTPClient:

    hostname: str
    port: int
    auth_type: AuthType
    key: Optional[bytes]
    hash: str
    authorized: bool
    reader: Optional[asyncio.StreamReader]
    writer: Optional[asyncio.StreamWriter]
    lock: asyncio.Lock

    def __init__(self, hostname: str = "127.0.0.1", port: int = 740, auth_type: AuthType = AuthType.Key, key: Optional[bytes | str] = None, hash: str = "") -> None:
        if not isinstance(hostname, str):
            raise TypeError("hostname should be instance of str")
        if not isinstance(port, int):
            raise TypeError("ip should be instance of int")
        if not isinstance(auth_type, AuthType):
            raise TypeError("auth_type should be instance of AuthType")
        if not isinstance(hash, str):
            raise TypeError("hash should be instance of str")
        if auth_type is AuthType.Key:
            if not key:
                raise TypeError("key is required with Key Auth")
            if not isinstance(key, bytes):
                if isinstance(key, str):
                    key = key.encode()
                else:
                    raise TypeError("key should be instance of bytes")
        
        self.hostname = hostname
        self.port = port
        self.auth_type = auth_type
        self.key = key
        self.hash = hash
        self.reader = None
        self.writer = None
        self.lock = asyncio.Lock()
        self.authorized = False

    async def auth(self) -> StatusCode:
        if self.authorized:
            raise AuthException("Already authorized")
        
        res = await self._send_packet(AuthPacket(self.auth_type, self.hash, self.key))
        if not res.ok:
            await self.close()
            if res.status_code == StatusCode.HashError:
                raise AuthException("Invalid procedure hash")
            raise AuthException(f"Authorization error with type {self.auth_type.name} ({res.status_code})")
        else:
            self.authorized = True
        return res.status_code

    async def send_procedure(self, name: int, subtype: int, pds: ProcedureDataStructure) -> ServerPacket:
        packet = ProcedurePacket(name, subtype, pds)
        return await self._send_packet(packet)

    async def _send_packet(self, packet: Packet) -> Optional[ServerPacket]:
        if not self.reader or not self.writer:
            raise UnopenedConnection
        if self.writer.is_closing():
            raise ClosedConnecton
        if packet.auth_required and not self.authorized:
            raise AuthException("Auth required before send this packet")
        
        async with self.lock:
            self.writer.write(packet.to_bytes())
            data = await self.reader.read(65535)
            if not data:
                raise ClosedConnecton("Server closed the connection")
            return ServerPacket.from_bytes(data)

    async def close(self) -> None:
        if not self.reader or not self.writer:
            raise UnopenedConnection
        
        await self._send_packet(ExitPacket())
        self.writer.close()
        await self.writer.wait_closed()

    async def connect(self) -> None:
        self.reader, self.writer = await asyncio.open_connection(self.hostname, self.port)
        self.authorized = False

    async def __aenter__(self) -> Self:
        await self.connect()
        await self.auth()

        return self
    
    async def __aexit__(self, exc_type, exc, tb) -> None:
        await self.close()


from .int import IntegerType
from .base import DataType, fieldtype


@fieldtype
class BooleanType(IntegerType):
    raw_type = 0xA2
    data_type = bool

    def to_bytes(self, data: DataType) -> bytes:
        if not isinstance(data, bool):
            raise TypeError("data should be instance of bool")
        return data.to_bytes()


from typing import Any
from lptp.server import manager
from datetime import datetime, UTC
from lptp.types.pds.types.base import FieldType


def hex_str(num: int) -> str:
    hex_data = hex(num)[2:]
    return "0x" + ("0" if len(hex_data) == 1 else "") + hex_data

def generate_proc_file(filename: str, manager: "manager.Manager") -> None:
    file_content = f"""
from lptp.client.client import LPTPClient
from lptp.exceptions.client import ProcedureError
from lptp.types.pds import DataField, ProcedureDataStructure
from typing import *

# This file autogenerated on {datetime.now(UTC).strftime('%d.%m.%Y %H:%M:%S')} UTC
# Total procedure count is {len(manager._procedures)}

class ProcedureManager:
    client: LPTPClient

    hash = "{manager.hash}"

    def __init__(self, client: LPTPClient = None) -> None:
        if client and not isinstance(client, LPTPClient):
            raise TypeError("client should be instance of LPTPClient")
        self.client = client
        self.client.hash = self.hash

    async def __aenter__(self) -> Self:
        await self.client.__aenter__()
        return self
    
    async def __aexit__(self, exc_type, exc, tb) -> None:
        return await self.client.__aexit__(exc_type, exc, tb)
"""
    for procedure in manager._procedures:
        params = ", ".join(["self"] + [
            f"{field.varname}: {field.name}%s" % (" = " + str(field.default) if field.have_default else "")
            for field in procedure.get_fields()
        ])
        fields = ",\n\t\t\t\t".join([
            f"DataField({hex_str(field.name)}, {field.varname})"
            for field in procedure.get_fields()
        ])
        file_content += f"""
    async def {procedure.get_func_name()}({params}) -> {procedure.get_return().type_name}:
        res = await self.client.send_procedure(
            {hex_str(procedure.name)},
            {hex_str(procedure.subtype)},
            ProcedureDataStructure([
                {fields}
            ])
        )
        if not res.ok:
            raise ProcedureError("Server returned %s status" % res.status_code)
        return res.field.data
"""
        
        with open(filename, "w") as f:
            f.write(file_content)


from abc import ABC, abstractmethod
from typing import Optional, Self

from lptp.types.pds.field import SingleDataField
from lptp import exceptions
from .status import StatusCode


class Packet(ABC):

    auth_required: bool = False

    @classmethod
    @abstractmethod
    def from_bytes(cls, data: bytes) -> Self:
        pass

    @abstractmethod
    def to_bytes(self) -> bytes:
        pass

class ServerPacket(Packet):
    status_code: StatusCode
    field: Optional[SingleDataField]

    def __init__(self, status_code: int, field: Optional[SingleDataField] = None) -> None:
        if not isinstance(status_code, int):
            raise TypeError("status_code should be instance of int")
        if field and not isinstance(field, SingleDataField):
            raise TypeError("field should be instance of DataField")

        self.status_code = status_code
        self.field = field

    def to_bytes(self) -> bytes:
        return self.status_code.to_bytes() + (self.field.to_bytes() if self.field else b'')
    
    def __repr__(self) -> str:
        status_code = self.status_code
        field = self.field
        return f"<{type(self).__name__} {status_code=} {field=}>"

    @property
    def ok(self) -> bool:
        return self.status_code is StatusCode.Ok

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
        
        if not data:
            raise exceptions.DataException("Cannot convert data")
        
        try:
            status_code = StatusCode(data[0])
        except ValueError:
            raise exceptions.DataException("Invalid status_code")
        if data[1:]:
            field = SingleDataField.from_bytes(data[1:])
        else:
            field = None
        
        return cls(status_code, field)


from typing import Callable, List, Optional
from .procedure import Procedure
import hashlib


class Manager:
    _procedures: List[Procedure]

    def __init__(self) -> None:
        self._procedures = []

    def procedure(self, coro, name: int = None, subtype: int = 0):
        procedure = Procedure(coro, name, subtype)
        self._procedures.append(procedure)
        self._reissue_names()
        return coro
    
    def sub_procedure(self, proc_func: Callable, subtype: int = 1):

        def inner(coro):
            procedure = self.get_by_func(proc_func)
            if not procedure:
                raise ValueError("Undefined procedure")
            return self.procedure(coro, procedure.name, subtype)
        return inner
    
    def _reissue_names(self):
        procedures = {
            procedure: list(filter(lambda p: p.name == procedure.name and p.subtype > 0, self._procedures))
            for procedure in self._procedures
            if procedure.subtype == 0
        }
        for i, (procedure, sub_procedures) in enumerate(sorted(procedures.items(), key=lambda p: p[0].get_func_name())):
            for proc in (procedure, *sub_procedures):
                proc.name = i

    def get_by_func(self, func: Callable) -> Optional[Procedure]:
        return next(filter(lambda proc: proc.coro == func, self._procedures), None)
    
    def get(self, name: int, subtype: int) -> Optional[Procedure]:
        return next(filter(lambda proc: proc.name == name and proc.subtype == subtype, self._procedures), None)
    
    def extend(self, manager: "Manager") -> None:
        if not isinstance(manager, Manager):
            raise TypeError("manager should be instance of Manager")
        self._procedures.extend(manager._procedures)
        self._reissue_names()

    @property
    def hash(self) -> str:
        procedures = b"".join([
            procedure.name.to_bytes() + procedure.subtype.to_bytes() + procedure.get_func_name().encode() + b"".join([
                field.name.to_bytes() + field.varname.encode() + str(field.field_type.data_type).encode() + str(field.default).encode()
                for field in procedure.get_fields()
            ])
            for procedure in self._procedures
        ])
        return hashlib.sha256(procedures).hexdigest()


import lptp.types


class RequestError(Exception):
    def __init__(self,  text: str = "", status_code: "lptp.types.StatusCode" = None) -> None:
        super().__init__(text)
        if not status_code: status_code = lptp.types.StatusCode.BadRequest
        self.status_code = status_code


from typing import List, Optional
from lptp.exceptions import RequestError
from lptp.server.file_generator import generate_proc_file
from lptp.types import AuthPacket, ExitPacket, ProcedurePacket, ServerPacket, AuthType, StatusCode
from lptp.types.pds.field import SingleDataField
from .manager import Manager
from .log import logger
import asyncio, traceback


class LPTPServer:
    hostname: str
    port: int
    connection_wait: int
    auth_type: AuthType
    key: Optional[bytes | str]
    ips: List[str]
    manager: Manager

    def __init__(
            self,
            hostname: str = "127.0.0.1",
            port: int = 740,
            auth_type: AuthType = AuthType.Key,
            key: Optional[bytes] = None,
            ips: List[str] = [],
            connection_wait: int = 30
        ) -> None:
        if not isinstance(hostname, str):
            raise TypeError("hostname should be instance of str")
        if not isinstance(port, int):
            raise TypeError("ip should be instance of int")
        if not isinstance(connection_wait, int):
            raise TypeError("connection_wait should be instance of int")
        if auth_type is AuthType.Key:
            if not key:
                raise TypeError("key is required with Key Auth")
            if not isinstance(key, bytes):
                if isinstance(key, str):
                    key = key.encode()
                else:
                    raise TypeError("key should be instance of bytes")
        if auth_type is AuthType.Ip:
            if not isinstance(ips, list):
                raise TypeError("ips should be instance of list")
        
        self.hostname = hostname
        self.port = port
        self.connection_wait = connection_wait
        self.auth_type = auth_type
        self.key = key
        self.ips = ips[:]
        self.manager = Manager()

    def add_manager(self, *args: List[Manager]) -> None:
        for manager in args:
            if not isinstance(manager, Manager):
                raise TypeError("manager children should be instance of Manager")
            self.manager.extend(manager)

    async def _run_forever(self) -> None:
        loop = asyncio.get_running_loop()
        self.server = await loop.create_server(
            lambda: LPTPProtocol(self),
            self.hostname,
            self.port
        )
        async with self.server:
            logger.info(f"Server started on {self.hostname}:{self.port}")
            await self.server.serve_forever()

    def generate_proc_file(self, filename: str = "lptp_procedures.py"):
        if not isinstance(filename, str):
            raise TypeError("filename should be instance of str")
        generate_proc_file(filename, self.manager)

    def run_forever(self) -> None:
        asyncio.run(self._run_forever())

class LPTPProtocol(asyncio.Protocol):
    server: LPTPServer
    transport: asyncio.WriteTransport
    ip: str
    port: int
    authorized: False
    timer: asyncio.Task

    def __init__(self, server: LPTPServer) -> None:
        self.server = server

    def connection_made(self, transport: asyncio.WriteTransport) -> None:
        self.transport = transport
        self.ip, self.port = transport.get_extra_info("peername")
        self.authorized = False
        self.timer = asyncio.ensure_future(self.connection_timer())

    def connection_lost(self, exc) -> None:
        logger.info(f"Closed connection from {self.ip}:{self.port}")

    async def connection_timer(self) -> None:
        await asyncio.sleep(self.server.connection_wait)
        self.transport.close()

    async def auth_handler(self, data: bytes) -> bool:
        packet = AuthPacket.from_bytes(data)
        if packet.hash != self.server.manager.hash:
            raise RequestError(status_code = StatusCode.HashError)
        if packet.type is AuthType.Key and self.server.auth_type is AuthType.Key:
            if packet.key != self.server.key:
                raise RequestError(status_code = StatusCode.AuthError)
        elif packet.type is AuthType.Ip and self.server.auth_type is AuthType.Ip:
            if self.ip not in self.server.ips:
                raise RequestError(status_code = StatusCode.AuthError)
        elif packet.type is not AuthType.Without or self.server.auth_type is not AuthType.Without:
            raise RequestError(status_code = AuthType.Without)
        self.transport.write(ServerPacket(StatusCode.Ok).to_bytes())
        self.authorized = True
        
    async def procedure_handler(self, data: bytes) -> None:
        packet = ProcedurePacket.from_bytes(data)
        procedure = self.server.manager.get(packet.name, packet.subtype)
        if not procedure:
            raise RequestError("Invalid procedure", StatusCode.Undefined)
        field = SingleDataField(await procedure.call(packet.pds))
        self.transport.write(ServerPacket(StatusCode.Ok, field).to_bytes())

    async def receive_handler(self, data: bytes) -> None:
        if data == ExitPacket().to_bytes():
            self.transport.write(ServerPacket(StatusCode.Ok).to_bytes())
            self.transport.close()
            return
        
        try:
            if not self.authorized:
                await self.auth_handler(data)
            else:
                await self.procedure_handler(data)
        except RequestError as exc:
            logger.debug(str(exc))
            self.transport.write(ServerPacket(exc.status_code).to_bytes())

    async def data_received_async(self, data: bytes) -> None:
        logger.info(f"New receive from {self.ip}:{self.port} with {len(data)} bytes")
        
        try:
            await self.receive_handler(data)
        except Exception as err:
            traceback.print_exception(err)
            self.transport.write(ServerPacket(StatusCode.InternalError).to_bytes())

        self.timer.cancel()
        self.timer = asyncio.ensure_future(self.connection_timer())

    def data_received(self, data: bytes):
        asyncio.ensure_future(self.data_received_async(data))


from typing import List
from lptp import exceptions
from .base import DataType, FieldType, fieldtype


@fieldtype
class ListType(FieldType[list]):
    raw_type = 0xA4
    data_type = list
    data_aliases = (List,)

    def to_bytes(self, data: DataType) -> bytes:
        if not isinstance(data, list):
            raise TypeError("data should be instance of int")
        res = b""
        for el in data:
            el_type = FieldType.by_datatype(type(el))
            if not el_type:
                raise TypeError(f"{type(el).__name__} is invalid FieldType")
            el_data = el_type.to_bytes(el)
            res += el_type.raw_type.to_bytes() + len(el_data).to_bytes(2) + el_data
        return res

    def from_bytes(self, data: bytes) -> DataType:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
        res = []
        while data:
            if len(data) < 3:
                raise exceptions.DataException("Cannot convert data")
            field_type = FieldType.by_rawtype(data[0])
            if not field_type:
                raise exceptions.DataException("Invalid field_type")
            content_length = int.from_bytes(data[1:3])
            if len(data) - 3 < content_length:
                raise exceptions.DataException("Field length exceeds data")
            res.append(field_type.from_bytes(data[3:content_length+3]))
            data = data[content_length+3:]
        return res


from .base import DataType, FieldType, fieldtype


@fieldtype
class StringType(FieldType[str]):
    raw_type = 0xA1
    data_type = str

    def to_bytes(self, data: DataType) -> bytes:
        if not isinstance(data, str):
            raise TypeError("data should be instance of str")
        return data.encode()
    
    def from_bytes(self, data: bytes) -> DataType:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
        return data.decode()


from typing import List, Self
from lptp import exceptions
from .field import DataField
from .types import FieldType


class ProcedureDataStructure:
    fields: List[DataField]

    def __init__(self, fields: List[DataField] = []) -> None:
        if not isinstance(fields, list):
            raise TypeError("fields should be list of DataField")
        for field in fields:
            if not isinstance(field, DataField):
                raise TypeError("fields children should be instance of DataField")
            if field.type.annotation:
                raise TypeError(f"{field.name} is only annotation")

        self.fields = fields[:]

    def add_field(self, field: DataField) -> None:
        if not isinstance(field, DataField):
            raise TypeError("data should be instance of DataField")
        
        self.fields.append(field)

    def get_params(self) -> list:
        return [field.data for field in sorted(self.fields, key=lambda f: f.name)]

    def to_bytes(self) -> bytes:
        data = b"".join([field.to_bytes() for field in self.fields])
        return data
    
    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")

        fields = []
        while data:
            if len(data) < 4:
                raise exceptions.DataException("Cannot convert data")
            name = data[0]
            field_type = FieldType.by_rawtype(data[1])
            if not field_type:
                raise exceptions.DataException("Invalid field_type")
            field_length = int.from_bytes(data[2:4])
            if len(data) - 4 < field_length:
                raise exceptions.DataException("Field length exceeds data")
            fields.append(DataField.from_bytes(name, data[4:field_length+4], field_type))
            data = data[4 + field_length:]
        return cls(fields)


from enum import Enum
from typing import Optional, Self
from lptp import exceptions
from .base import Packet


class AuthType(int, Enum):
    Without = 0x0A
    Key = 0x0B
    Ip = 0x0C

class AuthPacket(Packet):
    type: AuthType
    hash: bytes
    key: Optional[bytes]
    auth_required = False

    def __init__(self, type: AuthType, hash: str, key: Optional[bytes] = None) -> None:
        if not isinstance(type, AuthType):
            raise TypeError("type should be instance of AuthType")
        if not isinstance(hash, str):
            raise TypeError("hash should be instance of str")
        if key and not isinstance(key, bytes):
            if isinstance(key, str):
                key = key.encode()
            else:
                raise TypeError("key should be instance of bytes")
        
        self.type = type
        self.hash = hash
        self.key = key

    def to_bytes(self) -> bytes:
        hash = bytes.fromhex(self.hash)
        if len(hash) != 32:
            raise ValueError("hash size should be 32 bytes")
        return self.type.to_bytes() + hash + (self.key if self.key else b'')
    
    def __repr__(self) -> str:
        authtype = self.type
        hash = self.hash
        key = self.key
        return f"<{type(self).__name__} {authtype=} {key=} {hash=}>"

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
        
        if not data:
            raise exceptions.DataException("Cannot convert data")
        
        try:
            type = AuthType(data[0])
            hash = data[1:33]
            if type is AuthType.Key and not data[33:]:
                raise exceptions.DataException("Key is required for this AuthType")
        except ValueError:
            raise exceptions.DataException("Invalid auth_type")
        
        return cls(type, hash.hex(), data[33:])


from abc import ABC, abstractmethod
from typing import Generic, List, Optional, Tuple, TypeVar
from .. import pds

fieldtypes: List[type["FieldType"]] = []
DataType = TypeVar("DataType")


class FieldType(ABC, Generic[DataType]):
    raw_type: int
    data_type: DataType
    data_aliases: Tuple = tuple()
    annotation: bool = False
    real_type: Optional[DataType]

    def __init__(self, real_type: Optional[DataType] = None) -> None:
        self.type = real_type

    @abstractmethod
    def to_bytes(self, data: DataType) -> bytes:
        pass

    @abstractmethod
    def from_bytes(self, data: bytes) -> DataType:
        pass

    @staticmethod
    def by_datatype(data_type: type) -> Optional["FieldType"]:
        t = next(filter(lambda t: t.data_type == data_type or data_type in t.data_aliases, fieldtypes), None)
        return t(data_type) if t else None

    @staticmethod
    def by_rawtype(raw_type: int) -> Optional["FieldType"]:
        t = next(filter(lambda t: t.raw_type == raw_type, fieldtypes), None)
        return t() if t else None
    
    @property
    def name(self) -> str:
        return type(self).__name__
    
    @property
    def type_name(self) -> str:
        return self.data_type.__name__
    
    def validate_field(self, field: "pds.DataField") -> bool:
        return field.type == self

    def __eq__(self, __value: object) -> bool:
        return type(self) == type(__value)

    def __repr__(self) -> str:
        return self.name

def fieldtype(cls):
    fieldtypes.append(cls)
    return cls


from typing import Self
from lptp.types.pds import ProcedureDataStructure
from lptp import exceptions
from .base import Packet


class ProcedurePacket(Packet):
    name: int
    subtype: int
    pds: ProcedureDataStructure
    auth_required = True

    def __init__(self, name: int, subtype: int, pds: ProcedureDataStructure) -> None:
        if not isinstance(name, int):
            raise TypeError("name should be instance of int")
        if not isinstance(subtype, int):
            raise TypeError("subtype should be instance of int")
        if not isinstance(pds, ProcedureDataStructure):
            raise TypeError("PDS should be instance of ProcedureDataStructure")
            
        self.name = name
        self.subtype = subtype
        self.pds = pds

    def to_bytes(self) -> bytes:
        return self.name.to_bytes() + self.subtype.to_bytes() + self.pds.to_bytes()

    def __repr__(self) -> str:
        name = self.name
        subtype = self.subtype
        pds = self.pds
        return f"<{type(self).__name__} {name=} {subtype=} {pds=}>"

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
    
        if len(data) < 2:
            raise exceptions.DataException("Cannot convert data")

        name = data[0]
        subtype = data[1]
        if data[2:]:
            pds = ProcedureDataStructure.from_bytes(data[2:])
        else:
            pds = ProcedureDataStructure()
        return cls(name, subtype, pds)


from .base import DataType, FieldType, fieldtype
from lptp import exceptions
from .. import pds
from typing import Any, Optional, Union


class AnnotationType(FieldType[DataType]):
    raw_type = -1
    annotation = True

    def to_bytes(self, data: Any) -> bytes:
        raise NotImplementedError
    
    def from_bytes(self, data: bytes) -> Any:
        return NotImplementedError
    
    @property
    def type_name(self) -> str:
        return self.name

@fieldtype
class AnyType(AnnotationType[Any]):
    data_type = Any
    data_aliases = (type(Any),)
    
    def __eq__(self, __value: Any) -> bool:
        return True

    @property
    def name(self) -> str:
        return "Any"
    
# @fieldtype
# class UnionType(AnnotationType[Union]):
#     data_type = Union
#     data_aliases = (type(Union[Any]),)
#     args: list[FieldType]

#     def __init__(self, real_type: Union[Any]) -> None:
#         super().__init__(real_type)
#         self.args = []
#         for arg in real_type.__args__:
#             field_type = FieldType.by_datatype(type(arg))
#             if not field_type:
#                 raise TypeError(f"{field_type} is invalid FieldType")
#             self.args.append(field_type)

#     def validate_field(self, field: "pds.DataField") -> bool:
#         for arg in self.args:
#             if field.type == arg:
#                 return True
#         return False

#     @property
#     def name(self) -> str:
#         return f"Union[{', '.join([arg.type_name for arg in self.args])}]"

# @fieldtype
# class OptionalType(FieldType):
#     raw_type = 0xA3
#     data_type = Optional
#     data_aliases = (type(Optional),)
#     annonation = True

#     def to_bytes(self, data: Any) -> bytes:
#         raise NotImplementedError
    
#     def from_bytes(self, data: bytes) -> Any:
#         return NotImplementedError
    
#     def __eq__(self, __value: Any) -> bool:
#         return self

#     @property
#     def name(self) -> str:
#         return "Any"
    
#     @property
#     def type_name(self) -> str:
#         return self.name
    
@fieldtype
class NullType(FieldType[None]):
    raw_type = 0xA3
    data_type = None
    data_aliases = (type(None),)

    def to_bytes(self, data: DataType) -> bytes:
        return 0x00.to_bytes()
    
    def from_bytes(self, data: bytes) -> DataType:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
        if data != 0x00.to_bytes():
            raise exceptions.DataException("Invalid null byte")
        return None

    def __eq__(self, __value: object) -> bool:
        return __value is None
    
    @property
    def name(self) -> str:
        return "None"
    
    @property
    def type_name(self) -> str:
        return self.name


from setuptools import setup

def readme():
  with open('README.md', 'r', encoding="utf8") as f:
    return f.read()

setup(
    name='lptp',
    version='1.0.2',
    author='zoomdev',
    author_email='zoomdeveloper@yandex.ru',
    description='Данный протокол является протоколом RPC вида и служит для вызова процедур на удалённом сервере. Проект является исключительно учебным и не рекомендуется для реального использования.',
    long_description=readme(),
    long_description_content_type='text/markdown',
    package_dir={"": "src"},
    keywords='rpc',
    url="https://github.com/Zoom-Developer/LPTP",
    python_requires='>=3.10'
)


from typing import Self
from .base import Packet


class ExitPacket(Packet):
    auth_required = False

    def to_bytes(self) -> bytes:
        return 0x00.to_bytes()
    
    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        return cls()


from .base import DataType, FieldType, fieldtype


@fieldtype
class IntegerType(FieldType[int]):
    raw_type = 0xA0
    data_type = int

    def to_bytes(self, data: DataType) -> bytes:
        if not isinstance(data, int):
            raise TypeError("data should be instance of int")
        return data.to_bytes((data.bit_length() + 7) // 8, signed=True)
    
    def from_bytes(self, data: bytes) -> DataType:
        if not isinstance(data, bytes):
            raise TypeError("data should be instance of bytes")
        return int.from_bytes(data, signed=True)